from copy import deepcopy
from pprint import pprint
import tensorflow.keras as keras
from . import load
from . import blocks
from . import addons
import tfbox.utils.helpers as h


# 
# Model: Parent class for TBox models/blocks
# 
# a simple wrapper of keras.Model with the following additions:
#
#   - an optional classifier
#   - is_skip property 
#   - standardized naming for tfbox models/blocks  
# 
class Model(keras.Model):
    #
    # CONSTANTS
    #
    NAME='TFBoxModel'
    DEFAULT_KEY=NAME
    SEGMENT='segment'
    GLOBAL_POOLING='global_pooling'
    DEFAULT_CLASSIFIER=SEGMENT


    #
    # PUBLIC
    #
    def __init__(self,
            is_skip=False,  
            name=NAME,
            named_layers=True,
            noisy=True):
        super(Model, self).__init__()
        self.classifier=None
        self.is_skip=is_skip
        self.model_name=name
        self.named_layers=named_layers


    def set_classifier(self,
            nb_classes,
            config,
            classifier_names=None,
            group_maps=None,
            group_nb_classes=None,
            file_name='classifier',
            folder=load.TFBOX,
            from_logits=None):
        if nb_classes:
            if not isinstance(nb_classes,list):
                nb_classes=[nb_classes]
            if not classifier_names:
                classifier_names=[None]*len(nb_classes)
            self.classifier=[]
            for n,cname in zip(nb_classes,classifier_names):
                self.classifier.append(
                    self._get_classifier(
                        n,
                        config,
                        classifier_name=cname,
                        file_name=file_name,
                        folder=folder,
                        from_logits=from_logits))

        if group_maps:
            self.grouping=addons.Groups(group_maps)
            self.group_classifier=self._get_classifier(
                group_nb_classes or len(group_maps),
                config.get('group_classifier'),
                file_name=file_name,
                folder=folder,
                from_logits=from_logits)
        else:
            self.grouping=False


    def output(self,x):
        if self.classifier:
            x=[c(x) for c in self.classifier]
            if self.grouping:
                gx=self.grouping(x)
                if self.group_classifier:
                    gx=self.group_classifier(gx)
                x=+[gx]
        return x


    def layer_name(self,group=None,index=None):
        return blocks.layer_name(self.model_name,group,index=index,named=self.named_layers)


    #
    # INTERNAL
    #
    def _get_classifier(self,
            nb_classes,
            config,
            classifier_name=None,
            file_name='classifier',
            folder=load.TFBOX,
            from_logits=None):
        if config:
            config=deepcopy(config)
            config['name']=classifier_name or config.get('name')
            if from_logits in [True,False]:
                config=self._update_activation(config,from_logits)
            if nb_classes and config:
                if config is True:
                    config={}
                elif isinstance(config,str):
                    config={ 'classifier_type': config }
                else:
                    config=load.config(config,file_name,folder)
                classifier_type=config.pop( 'classifier_type', self.DEFAULT_CLASSIFIER )
                if classifier_type==Model.SEGMENT:
                    classifier=blocks.SegmentClassifier(
                        nb_classes=nb_classes,
                        **config)
                elif classifier_type==Model.GLOBAL_POOLING:
                    raise NotImplementedError('TODO: GAPClassifier')
                else:
                    raise NotImplementedError(f'{classifier_type} is not a valid classifier')
            else:
                classifier=False
            return classifier


    def _update_activation(self,config,from_logits):
        if config is True:
            config={}
        if config is False:
            config={
                'filters': False
            }
        config['output_act']=not from_logits
        return config






from . import base
from . import load
from .encoder import Encoder
from .decoder import Decoder
#
# Encoder: a flexible generic encoder
# 
class EncoderDecoder(base.Model):
    #
    # CONSTANTS
    #
    NAME='EncoderDecoder'
    def __init__(self,
            config,
            file_name=None,
            folder=load.TFBOX,
            nb_classes=None,
            group_maps=None,
            group_nb_classes=None,
            from_logits=None,
            add_classifier=False,
            name=NAME,
            named_layers=True,
            noisy=True):
        super(EncoderDecoder, self).__init__(
            name=name,
            named_layers=named_layers,
            noisy=noisy)
        self.config=load.config(
            config,
            file_name or EncoderDecoder.NAME,
            folder)
        if add_classifier is None:
            add_classifier=config.get('classifier',False)
        # encoder
        encoder_config=self.config['encoder']
        encoder_config=load.config(
            encoder_config,
            Encoder.NAME,
            folder)
        self.encoder=Encoder(
            encoder_config,
            return_empty_skips=True,
            add_classifier=False)
        # decoder
        decoder_config=self.config['decoder']
        decoder_config=load.config(
            decoder_config,
            Decoder.NAME,
            folder)
        self.decoder=Decoder(
            decoder_config,
            nb_classes=nb_classes,
            add_classifier=False)
        # classifier
        if add_classifier and nb_classes:
            self.set_classifier(
                nb_classes,
                self.config.get('classifier'),
                group_maps=group_maps,
                group_nb_classes=group_nb_classes,
                folder=folder,
                from_logits=from_logits)
        

    def __call__(self,inputs,training=False):
        x,skips=self.encoder(inputs)
        self.decoder.set_output(inputs)
        x=self.decoder(x,skips,training)
        return self.output(x)


    #
    # INTERNAL
    #
    def _value(self,value,default):
        if value is None:
            value=default
        return value





import tensorflow.keras.losses as losses
from .weighted import weighted_categorical_crossentropy, focal_cross_entropy, masked_binary_cross_entropy
#
# CONSTANTS
#
DEFAULT_LOSS='categorical_crossentropy'
DEFAULT_WEIGHTED_LOSS='weighted_categorical_crossentropy'


#
# LOSS FUNCTION DICT
#
LOSS_FUNCTIONS={
    'focal_cross_entropy': focal_cross_entropy,
    'masked_binary_cross_entropy': masked_binary_cross_entropy,
    'weighted_categorical_crossentropy': weighted_categorical_crossentropy,
    'categorical_crossentropy': losses.CategoricalCrossentropy
}


#
# MAIN
#
def get(loss_func=None,weights=None,**kwargs):
    if not loss_func:
        if weights:
            loss_func=DEFAULT_WEIGHTED_LOSS
        else:
            loss_func=DEFAULT_LOSS
    if weights:
        kwargs['weights']=weights
    if isinstance(loss_func,str):
        loss_func=LOSS_FUNCTIONS.get(loss_func,loss_func)
    if not isinstance(loss_func,str):
        loss_func=loss_func(**kwargs)
    return loss_func


def scaled_loss(loss_func,scale,**kwargs):
    if isinstance(loss_func,str):
        loss_func=get(loss_func=loss_func,**kwargs)
    def _loss(target,output):
        return scale*loss_func(target,output)
    return _loss


def scaled_losses(loss_funcs,scales,kwargs_list=None,**kwargs):
    scaled_losses=[]
    if not isinstance(loss_funcs,(tuple,list)):
        loss_funcs=[loss_funcs]*len(scales)
    for i,l in enumerate(loss_funcs):
        kw=kwargs.copy()
        if kwargs_list:
            kw.update(kwargs_list[i])
        scaled_losses.append(scaled_loss(l,scales[i],**kw))
    return scaled_losses



from pprint import pprint
from copy import deepcopy
from . import load
from . import base
from . import blocks


#
# CONSTANTS
#
DEFAULT_BTYPE='stack'
INT_BTYPE='conv'
STACK='stack'

#
# Encoder: a flexible generic encoder
# 
class Encoder(base.Model):
    #
    # CONSTANTS
    #
    NAME='Encoder'
    DEFAULT_CLASSIFIER=base.Model.GLOBAL_POOLING


    def __init__(self,
            config,
            file_name=None,
            folder=load.TFBOX,
            nb_classes=None,
            classifier_names=None,
            group_maps=None,
            group_nb_classes=None,
            from_logits=None,
            add_classifier=None,
            return_empty_skips=False,
            name=NAME,
            named_layers=True,
            noisy=True):
        super(Encoder, self).__init__(
            name=name,
            named_layers=named_layers,
            noisy=noisy)
        config=load.config(
            deepcopy(config),
            file_name or Encoder.NAME,
            folder)
        if add_classifier is None:
            add_classifier=config.get('classifier',False)
        blocks_config=config.get('blocks_config',[])
        self.stacked_blocks=[blocks.build_blocks(c) for c in blocks_config]
        self.return_empty_skips=return_empty_skips
        if add_classifier and nb_classes:
            self.set_classifier(
                nb_classes,
                config.get('classifier'),
                classifier_names=classifier_names,
                group_maps=group_maps,
                group_nb_classes=group_nb_classes,
                folder=folder,
                from_logits=from_logits)


    def __call__(self,inputs,training=False):
        x=inputs
        skips=[]
        for sb in self.stacked_blocks:
            for block in sb:
                #
                # SHOULD XIN HAVE BEEN PASSED TO UPDATE_SKIPS?
                #
                xin=x
                x=block(x)
                skips=self._update_skips(block,skips,x)
        if skips or self.return_empty_skips:
            return x, skips
        else:
            return self.output(x)


    #
    # INTERNAL
    #
    def _update_skips(self,block,skips,x,force_update=False):
        try:
            if force_update or (block.is_skip): 
                skips.append(x)
        except:
            pass
        return skips









